==> klayout-0.10.tar.gz: klayout-0.10/src/layCellTreeModel.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layCellTreeModel.h"
#include "layLayoutView.h"

#include <QtGui/QTreeView>

#include <string.h>

#include <algorithm>

namespace lay {

// --------------------------------------------------------------------
//  A compare functor for the cell tree items

struct cmp_cell_tree_items_f 
{
  bool operator() (const CellTreeItem *a, const CellTreeItem *b)
  {
    return a->less_than (b);
  }
};

// --------------------------------------------------------------------
//  A compare functor for the cell tree items vs. name

struct cmp_cell_tree_item_vs_name_f 
{
  bool operator() (const CellTreeItem *a, const char *name)
  {
    return a->less_than (name);
  }
};

// --------------------------------------------------------------------
//  CellTreeItem implementation

CellTreeItem::CellTreeItem (const db::Layout *layout, CellTreeItem *parent, db::CellGraph::cell_index_type cell_index, bool flat)
  : mp_layout (layout), mp_parent (parent), m_index (0), m_children (), m_cell_index (cell_index)
{
  db::CellGraph::cell_index_type ci = (db::CellGraph::cell_index_type) -1l;
  m_child_count = 0;

  if (! flat) {
    const db::Cell *cell = & mp_layout->cell (m_cell_index);
    for (db::Cell::const_iterator child = cell->begin (); child != cell->end (); ++child) {
      if (child->object ().cell_index () != ci) {
        ci = child->object ().cell_index ();
        ++m_child_count;
      }
    }
  }
}

CellTreeItem::~CellTreeItem ()
{
  for (std::vector<CellTreeItem *>::iterator c = m_children.begin (); c != m_children.end (); ++c) {
    delete *c;
  }
  m_children.clear ();
}

int 
CellTreeItem::children () const
{
  return m_child_count;
}

CellTreeItem *
CellTreeItem::child (int index) 
{
  if (int (m_children.size ()) <= index) {

    //  create a list of child sub-items
    //  this exploits the fact that the cell instances are sorted such that instances
    //  with equal cell index appear after another

    const db::Cell *cell = & mp_layout->cell (m_cell_index);
    db::CellGraph::cell_index_type ci = (db::CellGraph::cell_index_type) -1l;

    m_children.reserve (m_child_count);

    for (db::Cell::const_iterator child = cell->begin (); child != cell->end (); ++child) {
      if (child->object ().cell_index () != ci) {
        ci = child->object ().cell_index ();
        CellTreeItem *child_item = new CellTreeItem (mp_layout, this, ci, false);
        m_children.push_back (child_item);
      }
    }

    std::sort (m_children.begin (), m_children.end (), cmp_cell_tree_items_f ());

    for (size_t i = 0; i < m_children.size (); ++i) {
      m_children [i]->set_index (i);
    }

  }

  return m_children [index];
}

db::CellGraph::cell_index_type 
CellTreeItem::cell_index () const
{
  return m_cell_index;
}

CellTreeItem *
CellTreeItem::parent () const
{
  return mp_parent;
}

bool 
CellTreeItem::less_than (const CellTreeItem *b) const
{
  return strcmp (mp_layout->cell_name (m_cell_index), b->mp_layout->cell_name (b->m_cell_index)) < 0;
}

bool 
CellTreeItem::less_than (const char *name) const
{
  return strcmp (mp_layout->cell_name (m_cell_index), name) < 0;
}

// --------------------------------------------------------------------
//  CellTreeModel implementation

CellTreeModel::CellTreeModel (QTreeView *parent, lay::LayoutView *view, int cv_index, unsigned int flags, const db::Cell *base)
  : QAbstractItemModel (parent), 
    mp_tree_view (parent), mp_view (view), m_cv_index (cv_index)
{
  m_flat = ((flags & Flat) != 0);

  mp_layout = & view->cellview (cv_index)->layout ();

  if ((flags & Children) != 0) {

    m_flat = true; //  no "hierarchical children" yet.

    m_toplevel.reserve (base->child_cells ());

    for (db::Cell::child_cell_iterator child = base->begin_child_cells (); child != base->end_child_cells (); ++child) {
      CellTreeItem *item = new CellTreeItem (mp_layout, 0, *child, true);
      m_toplevel.push_back (item);
    }

  } else if ((flags & Parents) != 0) {

    m_flat = true; //  no "hierarchical parents" yet.

    m_toplevel.reserve (base->parent_cells ());

    for (db::Cell::parent_cell_iterator parent = base->begin_parent_cells (); parent != base->end_parent_cells (); ++parent) {
      CellTreeItem *item = new CellTreeItem (mp_layout, 0, *parent, true);
      m_toplevel.push_back (item);
    }

  } else {

    if (m_flat) {
      m_toplevel.reserve (mp_layout->cells ());
    }

    db::CellGraph::top_down_const_iterator top = mp_layout->begin_top_down ();
    while (top != mp_layout->end_top_down ()) {

      if (m_flat) {
        CellTreeItem *item = new CellTreeItem (mp_layout, 0, *top, true);
        m_toplevel.push_back (item);
      } else if (mp_layout->cell (*top).is_top ()) {
        CellTreeItem *item = new CellTreeItem (mp_layout, 0, *top, false);
        m_toplevel.push_back (item);
      } else {
        break;
      }

      ++top;

    }

  }

  std::sort (m_toplevel.begin (), m_toplevel.end (), cmp_cell_tree_items_f ());

  for (size_t i = 0; i < m_toplevel.size (); ++i) {
    m_toplevel [i]->set_index (i);
  }

}

CellTreeModel::~CellTreeModel ()
{
  for (std::vector<CellTreeItem *>::iterator c = m_toplevel.begin (); c != m_toplevel.end (); ++c) {
    delete *c;
  }
  m_toplevel.clear ();
}

Qt::ItemFlags 
CellTreeModel::flags (const QModelIndex &index) const
{
  return QAbstractItemModel::flags (index);
}

int 
CellTreeModel::columnCount (const QModelIndex &) const 
{
  return 1;
}

QVariant 
CellTreeModel::data (const QModelIndex &index, int role) const 
{
  CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
  if (role == Qt::DisplayRole || role == Qt::EditRole) {

    return QVariant (" " + QString (mp_layout->cell_name (item->cell_index ())) + " ");

  } else if (role == Qt::FontRole) {

    QFont f (mp_tree_view->font ());
    const lay::CellView::cell_path_type &path = mp_view->cellview (m_cv_index).path ();

    CellTreeItem *it = item;
    if (! path.empty ()) {
      lay::CellView::cell_path_type::const_iterator p = path.end ();
      while (it && p != path.begin ()) {
        --p;
        if (it->cell_index () != *p) {
          break;
        }
        it = it->parent ();
      }
      if (! it && p == path.begin ()) {
        f.setBold (true);
      }
    }

    if (mp_view->is_cell_hidden (item->cell_index (), m_cv_index)) {
      f.setStrikeOut (true);
    }

    return QVariant (f);

  } else if (role == Qt::TextColorRole) {

#if 0 // do strikeout rather than making the color darker
    QPalette pl (mp_tree_view->palette ());
    if (mp_view->is_cell_hidden (item->cell_index (), m_cv_index)) {
      QColor c1 = pl.color (QColorGroup::Text);
      QColor cb = pl.color (QColorGroup::Base);
      return QVariant (QColor ((c1.red () + cb.red ()) / 2, (c1.green () + cb.green ()) / 2, (c1.blue () + cb.blue ()) / 2));
    } else {
      return QVariant (pl.color (QColorGroup::Text));
    }
#else
    return QVariant ();
#endif

  } else {

    return QVariant ();

  }
}

QVariant 
CellTreeModel::headerData (int /*section*/, Qt::Orientation /*orientation*/, int /*role*/) const
{
  return QVariant ();
}

int 
CellTreeModel::rowCount (const QModelIndex &parent) const 
{
  if (parent.isValid ()) {
    CellTreeItem *item = (CellTreeItem *) parent.internalPointer ();
    return int (item->children ());
  } else {
    return int (m_toplevel.size ());
  }
}

QModelIndex 
CellTreeModel::index (int row, int column, const QModelIndex &parent) const 
{
  if (parent.isValid ()) {
    CellTreeItem *item = (CellTreeItem *) parent.internalPointer ();
    return createIndex (row, column, item->child (row));
  } else {
    return createIndex (row, column, m_toplevel [row]);
  }
}

QModelIndex 
CellTreeModel::parent (const QModelIndex &index) const 
{
  CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
  CellTreeItem *pitem = item->parent ();
  if (pitem) {
    return createIndex (int (pitem->index ()), index.column (), pitem);
  } else {
    return QModelIndex ();
  }
}

int 
CellTreeModel::toplevel_items () const
{
  return int (m_toplevel.size ());
}

CellTreeItem *
CellTreeModel::toplevel_item (int index) 
{
  return m_toplevel [index];
}

QModelIndex 
CellTreeModel::model_index (CellTreeItem *item) const
{
  return createIndex (int (item->index ()), 0, item);
}

db::CellGraph::cell_index_type 
CellTreeModel::cell_index (const QModelIndex &index) const
{
  CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
  return item->cell_index ();
}

const db::Cell *
CellTreeModel::cell (const QModelIndex &index) const
{
  if (index.isValid ()) {
    CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
    return & mp_layout->cell (item->cell_index ());
  } else {
    return 0;
  }
}

const char *
CellTreeModel::cell_name (const QModelIndex &index) const
{
  if (index.isValid ()) {
    CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
    return mp_layout->cell_name (item->cell_index ());
  } else {
    return 0;
  }
}

QModelIndex 
CellTreeModel::locate (const char *name) const
{
  std::vector <CellTreeItem *>::const_iterator lc = std::lower_bound (m_toplevel.begin (), m_toplevel.end (), name, cmp_cell_tree_item_vs_name_f ());
  
  if (lc == m_toplevel.end ()) {
    return QModelIndex (); 
  } else {
    return model_index (*lc);
  }
}

} // namespace lay

==> klayout-0.14.tar.gz: klayout-0.14/src/layCellTreeModel.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "layCellTreeModel.h"
#include "layLayoutView.h"

#include <QtGui/QTreeView>

#include <string.h>

#include <algorithm>

namespace lay {

// --------------------------------------------------------------------
//  A compare functor for the cell tree items

struct cmp_cell_tree_items_f 
{
  bool operator() (const CellTreeItem *a, const CellTreeItem *b)
  {
    return a->less_than (b);
  }
};

// --------------------------------------------------------------------
//  A compare functor for the cell tree items vs. name

struct cmp_cell_tree_item_vs_name_f 
{
  bool operator() (const CellTreeItem *a, const char *name)
  {
    return a->less_than (name);
  }
};

// --------------------------------------------------------------------
//  CellTreeItem implementation

CellTreeItem::CellTreeItem (const db::Layout *layout, CellTreeItem *parent, db::CellGraph::cell_index_type cell_index, bool flat)
  : mp_layout (layout), mp_parent (parent), m_index (0), m_children (), m_cell_index (cell_index)
{
  db::CellGraph::cell_index_type ci = (db::CellGraph::cell_index_type) -1l;
  m_child_count = 0;

  if (! flat) {
    const db::Cell *cell = & mp_layout->cell (m_cell_index);
    for (db::Cell::const_iterator child = cell->begin (); child != cell->end (); ++child) {
      if (child->object ().cell_index () != ci) {
        ci = child->object ().cell_index ();
        ++m_child_count;
      }
    }
  }
}

CellTreeItem::~CellTreeItem ()
{
  for (std::vector<CellTreeItem *>::iterator c = m_children.begin (); c != m_children.end (); ++c) {
    delete *c;
  }
  m_children.clear ();
}

int 
CellTreeItem::children () const
{
  return m_child_count;
}

CellTreeItem *
CellTreeItem::child (int index) 
{
  if (int (m_children.size ()) <= index) {

    //  create a list of child sub-items
    //  this exploits the fact that the cell instances are sorted such that instances
    //  with equal cell index appear after another

    const db::Cell *cell = & mp_layout->cell (m_cell_index);
    db::CellGraph::cell_index_type ci = (db::CellGraph::cell_index_type) -1l;

    m_children.reserve (m_child_count);

    for (db::Cell::const_iterator child = cell->begin (); child != cell->end (); ++child) {
      if (child->object ().cell_index () != ci) {
        ci = child->object ().cell_index ();
        CellTreeItem *child_item = new CellTreeItem (mp_layout, this, ci, false);
        m_children.push_back (child_item);
      }
    }

    std::sort (m_children.begin (), m_children.end (), cmp_cell_tree_items_f ());

    for (size_t i = 0; i < m_children.size (); ++i) {
      m_children [i]->set_index (i);
    }

  }

  return m_children [index];
}

db::CellGraph::cell_index_type 
CellTreeItem::cell_index () const
{
  return m_cell_index;
}

CellTreeItem *
CellTreeItem::parent () const
{
  return mp_parent;
}

bool 
CellTreeItem::less_than (const CellTreeItem *b) const
{
  return strcmp (mp_layout->cell_name (m_cell_index), b->mp_layout->cell_name (b->m_cell_index)) < 0;
}

bool 
CellTreeItem::less_than (const char *name) const
{
  return strcmp (mp_layout->cell_name (m_cell_index), name) < 0;
}

// --------------------------------------------------------------------
//  CellTreeModel implementation
//  Hint: it may happen that the cell tree model gets engaged while the layout is not
//  valid ("under construction"). In this case, the model will return defaults or void
//  objects.

CellTreeModel::CellTreeModel (QTreeView *parent, lay::LayoutView *view, int cv_index, unsigned int flags, const db::Cell *base)
  : QAbstractItemModel (parent), 
    mp_tree_view (parent), mp_view (view), m_cv_index (cv_index)
{
  mp_view->add_cell_visibility_observer (*this);
  mp_view->add_cellview_observer (*this);

  m_flat = ((flags & Flat) != 0);

  mp_layout = & view->cellview (cv_index)->layout ();
  tl_assert (! mp_layout->under_construction ());

  if ((flags & Children) != 0) {

    m_flat = true; //  no "hierarchical children" yet.

    m_toplevel.reserve (base->child_cells ());

    for (db::Cell::child_cell_iterator child = base->begin_child_cells (); child != base->end_child_cells (); ++child) {
      CellTreeItem *item = new CellTreeItem (mp_layout, 0, *child, true);
      m_toplevel.push_back (item);
    }

  } else if ((flags & Parents) != 0) {

    m_flat = true; //  no "hierarchical parents" yet.

    m_toplevel.reserve (base->parent_cells ());

    for (db::Cell::parent_cell_iterator parent = base->begin_parent_cells (); parent != base->end_parent_cells (); ++parent) {
      CellTreeItem *item = new CellTreeItem (mp_layout, 0, *parent, true);
      m_toplevel.push_back (item);
    }

  } else {

    if (m_flat) {
      m_toplevel.reserve (mp_layout->cells ());
    }

    db::CellGraph::top_down_const_iterator top = mp_layout->begin_top_down ();
    while (top != mp_layout->end_top_down ()) {

      if (m_flat) {
        CellTreeItem *item = new CellTreeItem (mp_layout, 0, *top, true);
        m_toplevel.push_back (item);
      } else if (mp_layout->cell (*top).is_top ()) {
        CellTreeItem *item = new CellTreeItem (mp_layout, 0, *top, false);
        m_toplevel.push_back (item);
      } else {
        break;
      }

      ++top;

    }

  }

  std::sort (m_toplevel.begin (), m_toplevel.end (), cmp_cell_tree_items_f ());

  for (size_t i = 0; i < m_toplevel.size (); ++i) {
    m_toplevel [i]->set_index (i);
  }

}

CellTreeModel::~CellTreeModel ()
{
  for (std::vector<CellTreeItem *>::iterator c = m_toplevel.begin (); c != m_toplevel.end (); ++c) {
    delete *c;
  }
  m_toplevel.clear ();
  mp_view->remove_cell_visibility_observer (*this);
}

Qt::ItemFlags 
CellTreeModel::flags (const QModelIndex &index) const
{
  return QAbstractItemModel::flags (index);
}

int 
CellTreeModel::columnCount (const QModelIndex &) const 
{
  return 1;
}

QVariant 
CellTreeModel::data (const QModelIndex &index, int role) const 
{
  CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
  if (! item || mp_layout->under_construction ()) {
    return QVariant ();
  }
  if (role == Qt::DisplayRole || role == Qt::EditRole) {

    return QVariant (" " + QString (mp_layout->cell_name (item->cell_index ())) + " ");

  } else if (role == Qt::FontRole) {

    QFont f (mp_tree_view->font ());
    const lay::CellView::unspecific_cell_path_type &path = mp_view->cellview (m_cv_index).unspecific_path ();
    const lay::CellView::specific_cell_path_type &ctx_path = mp_view->cellview (m_cv_index).specific_path ();

    if (! path.empty ()) {
      if (item->cell_index () == path.back ()) {
        if (m_flat) {
          f.setBold (true);
        } else {
          CellTreeItem *it = item;
          lay::CellView::unspecific_cell_path_type::const_iterator p = path.end ();
          while (it && p != path.begin ()) {
            --p;
            if (it->cell_index () != *p) {
              break;
            }
            it = it->parent ();
          }
          if (! it && p == path.begin ()) {
            f.setBold (true);
          }
        }
      } else if (! ctx_path.empty () && item->cell_index () == ctx_path.back ().cell_inst->object ().cell_index ()) {
        if (m_flat) {
          f.setUnderline (true);
        } else {
          CellTreeItem *it = item;
          lay::CellView::specific_cell_path_type::const_iterator cp = ctx_path.end ();
          while (it && cp != ctx_path.begin ()) {
            --cp;
            if (it->cell_index () != cp->cell_inst->object ().cell_index ()) {
              break;
            }
            it = it->parent ();
          }
          if (cp == ctx_path.begin ()) {
            lay::CellView::unspecific_cell_path_type::const_iterator p = path.end ();
            while (it && p != path.begin ()) {
              --p;
              if (it->cell_index () != *p) {
                break;
              }
              it = it->parent ();
            }
            if (! it && p == path.begin ()) {
              f.setUnderline (true);
            }
          }
        }
      }
    }

    if (mp_view->is_cell_hidden (item->cell_index (), m_cv_index)) {
      f.setStrikeOut (true);
    }

    return QVariant (f);

  } else if (role == Qt::TextColorRole) {

#if 0 // do strikeout rather than making the color darker
    QPalette pl (mp_tree_view->palette ());
    if (mp_view->is_cell_hidden (item->cell_index (), m_cv_index)) {
      QColor c1 = pl.color (QColorGroup::Text);
      QColor cb = pl.color (QColorGroup::Base);
      return QVariant (QColor ((c1.red () + cb.red ()) / 2, (c1.green () + cb.green ()) / 2, (c1.blue () + cb.blue ()) / 2));
    } else {
      return QVariant (pl.color (QColorGroup::Text));
    }
#else
    return QVariant ();
#endif

  } else {

    return QVariant ();

  }
}

QVariant 
CellTreeModel::headerData (int /*section*/, Qt::Orientation /*orientation*/, int /*role*/) const
{
  return QVariant ();
}

int 
CellTreeModel::rowCount (const QModelIndex &parent) const 
{
  if (mp_layout->under_construction ()) {
    return 0;
  } else if (parent.isValid ()) {
    CellTreeItem *item = (CellTreeItem *) parent.internalPointer ();
    if (! item) {
      return 0;
    } else {
      return int (item->children ());
    }
  } else {
    return int (m_toplevel.size ());
  }
}

QModelIndex 
CellTreeModel::index (int row, int column, const QModelIndex &parent) const 
{
  if (mp_layout->under_construction ()) {
    return QModelIndex ();
  } else if (parent.isValid ()) {
    CellTreeItem *item = (CellTreeItem *) parent.internalPointer ();
    if (! item) {
      return QModelIndex ();
    } else {
      return createIndex (row, column, item->child (row));
    }
  } else if (row >= 0 && row < int (m_toplevel.size ())) {
    return createIndex (row, column, m_toplevel [row]);
  } else {
    return QModelIndex ();
  }
}

QModelIndex 
CellTreeModel::parent (const QModelIndex &index) const 
{
  if (mp_layout->under_construction ()) {
    return QModelIndex ();
  } 
  if (! index.isValid ()) {
    return index;
  }
  CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
  if (! item) {
    return QModelIndex ();
  }
  CellTreeItem *pitem = item->parent ();
  if (pitem) {
    return createIndex (int (pitem->index ()), index.column (), pitem);
  } else {
    return QModelIndex ();
  }
}

int 
CellTreeModel::toplevel_items () const
{
  if (mp_layout->under_construction ()) {
    return 0;
  } else {
    return int (m_toplevel.size ());
  }
}

CellTreeItem *
CellTreeModel::toplevel_item (int index) 
{
  if (mp_layout->under_construction ()) {
    return 0;
  } else {
    return m_toplevel [index];
  }
}

QModelIndex 
CellTreeModel::model_index (CellTreeItem *item) const
{
  if (mp_layout->under_construction ()) {
    return QModelIndex ();
  } else {
    return createIndex (int (item->index ()), 0, item);
  }
}

db::CellGraph::cell_index_type 
CellTreeModel::cell_index (const QModelIndex &index) const
{
  if (mp_layout->under_construction ()) {
    return 0;
  } else {
    CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
    return item->cell_index ();
  }
}

const db::Cell *
CellTreeModel::cell (const QModelIndex &index) const
{
  if (index.isValid () && ! mp_layout->under_construction ()) {
    CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
    return & mp_layout->cell (item->cell_index ());
  } else {
    return 0;
  }
}

const char *
CellTreeModel::cell_name (const QModelIndex &index) const
{
  if (index.isValid () && ! mp_layout->under_construction ()) {
    CellTreeItem *item = (CellTreeItem *) index.internalPointer ();
    return mp_layout->cell_name (item->cell_index ());
  } else {
    return 0;
  }
}

QModelIndex 
CellTreeModel::locate (const char *name) const
{
  if (mp_layout->under_construction ()) {
    return QModelIndex ();
  }

  std::vector <CellTreeItem *>::const_iterator lc = std::lower_bound (m_toplevel.begin (), m_toplevel.end (), name, cmp_cell_tree_item_vs_name_f ());
  
  if (lc == m_toplevel.end ()) {
    return QModelIndex (); 
  } else {
    return model_index (*lc);
  }
}

QModelIndex 
CellTreeModel::topLeft () const
{
  if (m_toplevel.empty () || mp_layout->under_construction ()) {
    return QModelIndex ();
  } else {
    return model_index (m_toplevel.front ());
  }
}

QModelIndex 
CellTreeModel::bottomRight () const
{
  if (m_toplevel.empty () || mp_layout->under_construction ()) {
    return QModelIndex ();
  } else {
    QModelIndex p = model_index (m_toplevel.back ());
    int nr = 0;
    while (p.isValid () && (nr = rowCount (p)) > 0) {
      p = index (nr - 1, 0, p);
    }
    return p;
  }
}

void
CellTreeModel::signal_observed ()
{
  signal_data_changed ();
}

} // namespace lay
